/*
 * @(#)JNIAltivec.java created 12/20/2002 Berkeley, Office 405
 *
 * Copyright (c) 2002 Luca Lutterotti All Rights Reserved.
 *
 * This software is the research result of Luca Lutterotti and it is 
 * provided as it is as confidential and proprietary information.  
 * You shall not disclose such Confidential Information and shall use 
 * it only in accordance with the terms of the license agreement you 
 * entered into with the author.
 *
 * THE AUTHOR MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE SUITABILITY OF THE
 * SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
 * PURPOSE, OR NON-INFRINGEMENT. THE AUTHOR SHALL NOT BE LIABLE FOR ANY DAMAGES
 * SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING OR DISTRIBUTING
 * THIS SOFTWARE OR ITS DERIVATIVES.
 *
 */
 
package it.unitn.maud;

import java.util.*;

/**
 * The JNIAltivec is a interface library to the altivec native library
 * at the present only the routines in vecLib/vfp.h have been interfaced
 * natively. Check troughout this file to see what has been implemented.
 * The original comment of the vecLib/vfp.h file from Apple have been
 * maintained.
 * For some routines v******, corresponding s****** routines have been constructed
 * for pure native (not altivec) computing, for completeness.
 * The altivec methods start with v*****.
 * Remember to run 'make test' from the command line to check the speed of the 
 * altivec computation. You will see that the performances, compared to pure native
 * or pure java, depend on the size of your array. The best size for speed may depends on
 * your cache and altivec type, G4 processor type. So check for your computer.
 * The pure native computation is most of the time slower than pure java due to the
 * overhead of the JNI system that is not convenient for short computations like
 * calling just a mathematical function.
 * For my test (powerbook G4/800 MHz), the best size is always comprised between
 * 256 and 512.
 *
 * New release: 0.5 (thanks to Andy Belk from Apple for the suggestions to improve it)
 * The calling process has been unified. The library can be used also on other platforms.
 * Simply will use native computation instead of vector one.
 * So have a look at all the functions actually implemented.
 *
 * To use the package use the following import:
 *
 * import it.unitn.maud.JNIAltivec;
 * 
 * Then in your code just write:
 *
 * 												JNIAltivec.cos(A, B);
 *
 * where A is an array of float (whatever size) with the cos argument and the result is returned 
 * in the array of float B.
 * For speed optimization is better that also B is a non null array, otherwise the library will
 * inizialize it at speed expenses. If the library and altivec is available the computation wil be 
 * in the vector processor. If only the library is available it will be native, otherwise in
 * pure java (note: at the present some of the fancy mathematical functions don't have a
 * pure java correspective and have not been implemented in pure java).
 *
 * The library no more requires arrays multiple of 4.
 *
 * You can also call directly the native routines.
 *
 * @version $Revision: 0.5 $, $Date: 2003/02/06 09:22:23 $
 * @author Luca Lutterotti
 * @since JDK1.3
 */

public class JNIAltivec {

		static final short NO_LIBRARY = 0;
		static final short NATIVE_LIBRARY = 1;
		static final short ALTIVEC_LIBRARY = 2;

		static boolean altivecLibraryAvailable = false;
		static boolean nativeLibraryLoaded = false;
		static short libraryType = NO_LIBRARY;

    public static final boolean loadLibrary() {
			// for compatibility only, replaced by the static loading
			return nativeLibraryLoaded;
		}
		
		public static void forceScalarNativeComputation() {
			try {
				boolean trial = IsAltiVecAvailable();
				altivecLibraryAvailable = false;
				nativeLibraryLoaded = true;
			} catch (Throwable e) {
				System.out.println("Altivec native library not loaded!");
				e.printStackTrace();
			}
		}
		
		public static void forceVectorNativeComputation() {
			try {
				altivecLibraryAvailable = IsAltiVecAvailable();
				nativeLibraryLoaded = true;
			} catch (Throwable e) {
				System.out.println("Altivec native library not loaded!");
				e.printStackTrace();
			}
		}
		
		public static void forcePureJavaComputation() {
			altivecLibraryAvailable = false;
			nativeLibraryLoaded = false;
		}
		
		static {
        // Ensure native JNI library is loaded
//			System.out.println("Trying to load the altivec native library");
    	try {
        System.loadLibrary("JNIAltivec");
				System.out.println("G4 altivec library loaded!");
        nativeLibraryLoaded = true;
				libraryType = NATIVE_LIBRARY;
				if (altivecLibraryAvailable = IsAltiVecAvailable())
					libraryType = ALTIVEC_LIBRARY;
      } catch (Throwable e) {
      	e.printStackTrace();
      	System.out.println("Failed to load the G4 altivec library (only Mac OS X); will not use altivec acceleration!");
      }
    }

    public static final boolean isAltiVecAvailable() {
			return altivecLibraryAvailable;
    }
    
    public static final boolean isNativeComputationAvailable() {
			return nativeLibraryLoaded;
    }
    
    public static native boolean IsAltiVecAvailable(); 


/*
     File:       vecLib/vfp.h
 
     Contains:   MathLib style functions for vectors
 
     Version:    vecLib-138.1~1
 
     Copyright:   1999-2002 by Apple Computer, Inc., all rights reserved.
 
     Bugs?:      For bug reports, consult the following page on
                 the World Wide Web:
 
                     http://developer.apple.com/bugreporter/
 
*/

/*

                                                                                
    A collection of numerical functions designed to facilitate a wide         
    range of numerical programming for the Altivec Programming model.       
                                                                                

*/

/*
[ Computational Functions]
                                                                                
    vdivf        C = A  B                                                          
    vsqrtf       B = A                                                         
    vrsqrtf      B = 1/A                                                       
                                                                                

*/
/*
 *  vdivf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */  
 // safe method, should work also for arrays not multiple of 4 and different size, C can be null
 // return 0 if correctly executed, 1 if A or B are null
  public static native void vdivf(float[] A, float[] B, float[] C);
 // unsafe method, requires arrays multiple of 4 and same size, C cannot be null, may be is faster
  public static native void vdivfU(float[] A, float[] B, float[] C);
 // native not altivec
  public static native void sdivf(float[] A, float[] B, float[] C);
 // Altivec original from Andy Belt, can be faster if java floats are aligned, but it never happen
 // may be in the future?
  public static native int vdivf2(float[] A, float[] B, float[] C);
 // my Altivec original modified by Andy Belt, the fastest Altivec, requires all aligned
  public static native int vdivfF(float[] A, float[] B, float[] C);
	
// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void divf(float[] A, float[] B, float[] C) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null || B == null)
					break;
				int size = Math.min(A.length, B.length);
				if (C == null || (C.length < size))
					C = new float[size];
				for (int i = 0; i < size; i++)
					C[i] = A[i] / B[i];
			break;
			case NATIVE_LIBRARY:
				sdivf(A, B, C);
			break;
			case ALTIVEC_LIBRARY:
				vdivf(A, B, C);
			break;
		}
	}

/*
 *  vsqrtf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void vsqrtf(float[] A, float[] B);
  public static native void ssqrtf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void sqrtf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.sqrt(A[i]);
			break;
			case NATIVE_LIBRARY:
				ssqrtf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vsqrtf(A, B);
			break;
		}
	}


/*
 *  vrsqrtf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vrsqrtf(float[] A, float[] B);
  public static native void  srsqrtf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void rsqrtf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = 1f / (float) Math.sqrt(A[i]);
			break;
			case NATIVE_LIBRARY:
				srsqrtf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vrsqrtf(A, B);
			break;
		}
	}


/*
[ Exponential Functions]
                                                                                
    vexpf       B = Exp(A)                                                      
    vexpm1f     ExpM1(x) = Exp(x) - 1.  But, for small enough arguments,        
                ExpM1(x) is expected to be more accurate than Exp(x) - 1.       
    vlogf       B = Log(A)                                                      
    vlog1pf     Log1P = Log(1 + x). But, for small enough arguments,            
                Log1P is expected to be more accurate than Log(1 + x).          
    vlogbf      Extracts the exponent of its argument, as a signed integral     
                value. A subnormal argument is treated as though it were first  
                normalized. Thus:                                               
                1   <=   x * 2^(-logb(x))   <   2                           
    vscalbf     Computes x * 2^n efficently.  This is not normally done by      
                computing 2^n explicitly.                                       
                                                                                

*/
/*
 *  vexpf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vexpf(float[] A, float[] B);
  public static native void  sexpf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void expf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.exp(A[i]);
			break;
			case NATIVE_LIBRARY:
				sexpf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vexpf(A, B);
			break;
		}
	}

/*
 *  vexpm1f()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vexpm1f(float[] A, float[] B);
  public static native void  sexpm1f(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void expm1f(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.exp(A[i]) - 1f;
			break;
			case NATIVE_LIBRARY:
				sexpm1f(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vexpm1f(A, B);
			break;
		}
	}

/*
 *  vlogf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vlogf(float[] A, float[] B);
  public static native void  slogf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void logf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.log(A[i]);
			break;
			case NATIVE_LIBRARY:
				slogf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vlogf(A, B);
			break;
		}
	}

/*
 *  vlog1pf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vlog1pf(float[] A, float[] B);
  public static native void  slog1pf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void log1pf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.log(1.0 + A[i]);
			break;
			case NATIVE_LIBRARY:
				slog1pf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vlog1pf(A, B);
			break;
		}
	}

/*
 *  vlogbf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vlogbf(float[] A, float[] B);
  public static native void  slogbf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void logbf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
			// no support for now in Java
			System.out.println("No pure Java computation available for logb");
/*				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.logb(A[i]); */
			break;
			case NATIVE_LIBRARY:
				slogbf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vlogbf(A, B);
			break;
		}
	}

/*
 *  vscalbf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vscalbf(float[] A, int[] n, float[] B);
  public static native void  sscalbf(float[] A, int[] n, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void scalbf(float[] A, int[] n, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
			// no support for now in Java
			System.out.println("No pure Java computation available for scalb");
/*				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.scalb(A[i], n[i]); */
			break;
			case NATIVE_LIBRARY:
				sscalbf(A, n, B);
			break;
			case ALTIVEC_LIBRARY:
				vscalbf(A, n, B);
			break;
		}
	}

/*
[ Auxiliary Functions]
                                                                                
    vfabf           Absolute value is part of the programming model, however    
                    completeness it is included in the library.                 
    vcopysignf      Produces a value with the magnitude of its first argument   
                    and sign of its second argument.  NOTE: the order of the    
                    arguments matches the recommendation of the IEEE 754        
                    floating point standard,  which is opposite from the SANE   
                    copysign function.                                          
    vnextafterf     Computes the next representable value after 'A' in the      
                    direction of 'B'.  if A == B, then B is returned.           
                                                                                

*/
/*
 *  vfabf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vfabf(float[] A, float[] B);
  public static native void  sfabf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void fabf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
				System.out.println("Warning: function fab non available in pure java");
/*				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.fab(A[i]);*/
			break;
			case NATIVE_LIBRARY:
				sfabf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vfabf(A, B);
			break;
		}
	}


/*
 *  vcopysignf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vcopysignf(float[] A, float[] B, float[] C);
  public static native void  scopysignf(float[] A, float[] B, float[] C);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void copysignf(float[] A, float[] B, float[] C) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null || B == null)
					break;
				int size = Math.min(A.length, B.length);
				if (C == null || (C.length < size))
					C = new float[size];
				for (int i = 0; i < size; i++)
					C[i] = A[i] * (B[i]>=0?1:-1);
			break;
			case NATIVE_LIBRARY:
				scopysignf(A, B, C);
			break;
			case ALTIVEC_LIBRARY:
				vcopysignf(A, B, C);
			break;
		}
	}


/*
 *  vnextafterf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vnextafterf(float[] A, float[] B, float[] C);
  public static native void  snextafterf(float[] A, float[] B, float[] C);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void nextafterf(float[] A, float[] B, float[] C) {
		switch (libraryType) {
			case NO_LIBRARY:
			// no support for now in Java
			System.out.println("No pure Java computation available for nextafter");
/*				if (A == null || B == null)
					break;
				int size = Math.min(A.length, B.length);
				if (C == null || (C.length < size))
					C = new float[size];
				for (int i = 0; i < size; i++)
					C[i] = Math.nextafter(A[i], B[i]); */
			break;
			case NATIVE_LIBRARY:
				snextafterf(A, B, C);
			break;
			case ALTIVEC_LIBRARY:
				vnextafterf(A, B, C);
			break;
		}
	}


/*
[ Inquiry Functions]
                                                                                
    vclassifyf      Returns one of the FP_ values.                             
    vsignbitf       Non-zero if and only if the sign of the argument x is       
                    negative.  This includes, NaNs, infinities and zeros.       
                                                                                

*/
/*
 *  vclassifyf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vclassifyf(float[] A, int[] n);
  public static native void  sclassifyf(float[] A, int[] n);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void classifyf(float[] A, int[] n) {
		switch (libraryType) {
			case NO_LIBRARY:
			// no support for now in Java
			System.out.println("No pure Java computation available for classify");
/*				if (A == null)
					break;
				int size = A.length;
				if (n == null || (n.length < size))
					n = new int[size];
				for (int i = 0; i < size; i++)
					n[i] = Math.classify(A[i]); */
			break;
			case NATIVE_LIBRARY:
				sclassifyf(A, n);
			break;
			case ALTIVEC_LIBRARY:
				vclassifyf(A, n);
			break;
		}
	}


/*
 *  vsignbitf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vsignbitf(float[] A, int[] n);
  public static native void  ssignbitf(float[] A, int[] n);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void signbitf(float[] A, int[] n) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null)
					break;
				int size = A.length;
				if (n == null || (n.length < size))
					n = new int[size];
				for (int i = 0; i < size; i++)
					n[i] = A[i] < 0?1:0;
			break;
			case NATIVE_LIBRARY:
				ssignbitf(A, n);
			break;
			case ALTIVEC_LIBRARY:
				vsignbitf(A, n);
			break;
		}
	}


/*
[ Transcendental Functions]
                                                                                
    vsinf           B = Sin(A).                                                 
    vcosf           B = Cos(A).                                                 
    vtanf           B = Tan(A).                                                 
                                                                                

*/
/*
 *  vsinf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vsinf(float[] A, float[] B);
  public static native void  ssinf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void sinf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.sin(A[i]);
			break;
			case NATIVE_LIBRARY:
				ssinf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vsinf(A, B);
			break;
		}
	}


/*
 *  vcosf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vcosf(float[] A, float[] B);
  public static native void  scosf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void cosf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.cos(A[i]);
			break;
			case NATIVE_LIBRARY:
				scosf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vcosf(A, B);
			break;
		}
	}


/*
 *  vtanf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vtanf(float[] A, float[] B);
  public static native void  stanf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void tanf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.tan(A[i]);
			break;
			case NATIVE_LIBRARY:
				stanf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vtanf(A, B);
			break;
		}
	}


/*
[ Trigonometric Functions]
                                                                                
    vasinf      result is in [-pi/2,pi/2].                                      
    vacosf      result is in [0,pi].                                            
    vatanf      result is in [-pi/2,pi/2].                                      
    vatan2f     Computes the arc tangent of B/A in [-pi,pi] using the sign of   
                both arguments to determine the quadrant of the computed value. 
                                                                                

*/
/*
 *  vasinf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vasinf(float[] A, float[] B);
  public static native void  sasinf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void asinf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.asin(A[i]);
			break;
			case NATIVE_LIBRARY:
				sasinf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vasinf(A, B);
			break;
		}
	}


/*
 *  vacosf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vacosf(float[] A, float[] B);
  public static native void  sacosf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void acosf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.acos(A[i]);
			break;
			case NATIVE_LIBRARY:
				sacosf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vacosf(A, B);
			break;
		}
	}


/*
 *  vatanf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vatanf(float[] A, float[] B);
  public static native void  satanf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void atanf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.atan(A[i]);
			break;
			case NATIVE_LIBRARY:
				satanf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vatanf(A, B);
			break;
		}
	}


/*
 *  vatan2f()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vatan2f(float[] A, float[] B, float[] C);
  public static native void  satan2f(float[] A, float[] B, float[] C);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void atan2f(float[] A, float[] B, float[] C) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null || B == null)
					break;
				int size = Math.min(A.length, B.length);
				if (C == null || (C.length < size))
					C = new float[size];
				for (int i = 0; i < size; i++)
					C[i] = (float) Math.atan2(A[i], B[i]);
			break;
			case NATIVE_LIBRARY:
				satan2f(A, B, C);
			break;
			case ALTIVEC_LIBRARY:
				vatan2f(A, B, C);
			break;
		}
	}

/*
[ Hyperbolic Functions]
                                                                                
    vsinhf      Sine Hyperbolic.                                                
    vcoshf      Cosine Hyperbolic.                                              
    vtanhf      Tangent Hyperbolic.                                             
    vasinhf     Arcsine Hyperbolic.
    vacoshf     Arccosine Hyperbolic.
    vatanhf     Atctangent Hyperbolic.
                                                                                

*/
/*
 *  vsinhf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vsinhf(float[] A, float[] B);
  public static native void  ssinhf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void sinhf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
			System.out.println("No pure Java computation available for sinh");
/*				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.sinh(A[i]);*/
			break;
			case NATIVE_LIBRARY:
				ssinhf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vsinhf(A, B);
			break;
		}
	}


/*
 *  vcoshf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vcoshf(float[] A, float[] B);
  public static native void  scoshf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void coshf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
			System.out.println("No pure Java computation available for cosh");
/*				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.cosh(A[i]);*/
			break;
			case NATIVE_LIBRARY:
				scoshf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vcoshf(A, B);
			break;
		}
	}


/*
 *  vtanhf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vtanhf(float[] A, float[] B);
  public static native void  stanhf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void tanhf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
			System.out.println("No pure Java computation available for tanh");
/*				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.tanh(A[i]);*/
			break;
			case NATIVE_LIBRARY:
				stanhf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vtanhf(A, B);
			break;
		}
	}


/*
 *  vasinhf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vasinhf(float[] A, float[] B);
  public static native void  sasinhf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void asinhf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
			System.out.println("No pure Java computation available for asinh");
/*				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.asinh(A[i]);*/
			break;
			case NATIVE_LIBRARY:
				sasinhf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vasinhf(A, B);
			break;
		}
	}


/*
 *  vacoshf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vacoshf(float[] A, float[] B);
  public static native void  sacoshf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void acoshf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
			System.out.println("No pure Java computation available for acosh");
/*				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.acosh(A[i]);*/
			break;
			case NATIVE_LIBRARY:
				sacoshf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vacoshf(A, B);
			break;
		}
	}


/*
 *  vatanhf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vatanhf(float[] A, float[] B);
  public static native void  satanhf(float[] A, float[] B);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void atanhf(float[] A, float[] B) {
		switch (libraryType) {
			case NO_LIBRARY:
			System.out.println("No pure Java computation available for atanh");
/*				if (A == null)
					break;
				int size = A.length;
				if (B == null || (B.length < size))
					B = new float[size];
				for (int i = 0; i < size; i++)
					B[i] = (float) Math.atanh(A[i]);*/
			break;
			case NATIVE_LIBRARY:
				satanhf(A, B);
			break;
			case ALTIVEC_LIBRARY:
				vatanhf(A, B);
			break;
		}
	}


/*
[ Remainder Functions]
                                                                                
    vfmodf          B = X mod Y.                                                
    vremainderf     IEEE 754 floating point standard for remainder.             
    vremquof        SANE remainder.  It stores into 'quotient' the 7 low-order  
                    bits of the integer quotient x/y, such that:                
                    -127 <= quotient <= 127.                                
                                                                                

*/
/*
 *  vfmodf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vfmodf(float[] A, float[] B, float[] C);
  public static native void  sfmodf(float[] A, float[] B, float[] C);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void fmodf(float[] A, float[] B, float[] C) {
		switch (libraryType) {
			case NO_LIBRARY:
			System.out.println("No pure Java computation available for fmod");
/*				if (A == null || B == null)
					break;
				int size = Math.min(A.length, B.length);
				if (C == null || (C.length < size))
					C = new float[size];
				for (int i = 0; i < size; i++)
					C[i] = Math.fmod(A[i], B[i]);*/
			break;
			case NATIVE_LIBRARY:
				sfmodf(A, B, C);
			break;
			case ALTIVEC_LIBRARY:
				vfmodf(A, B, C);
			break;
		}
	}

/*
 *  vremainderf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vremainderf(float[] A, float[] B, float[] C);
  public static native void  sremainderf(float[] A, float[] B, float[] C);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void remainderf(float[] A, float[] B, float[] C) {
		switch (libraryType) {
			case NO_LIBRARY:
			System.out.println("No pure Java computation available for remainder");
/*				if (A == null || B == null)
					break;
				int size = Math.min(A.length, B.length);
				if (C == null || (C.length < size))
					C = new float[size];
				for (int i = 0; i < size; i++)
					C[i] = Math.remainder(A[i], B[i]);*/
			break;
			case NATIVE_LIBRARY:
				sremainderf(A, B, C);
			break;
			case ALTIVEC_LIBRARY:
				vremainderf(A, B, C);
			break;
		}
	}

/*
 *  vremquof()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vremquof(float[] A, float[] B, int[] QUO, float[] C);
  public static native void  sremquof(float[] A, float[] B, int[] QUO, float[] C);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void remquof(float[] A, float[] B, int[] QUO, float[] C) {
		switch (libraryType) {
			case NO_LIBRARY:
				System.out.println("Warning: no remquo function available in pure java");
/*				if (A == null || B == null)
					break;
				int size = Math.min(A.length, B.length);
				if (C == null || (C.length < size))
					C = new float[size];
				for (int i = 0; i < size; i++)
					C[i] = Math.remquo(A[i], B[i]);*/
			break;
			case NATIVE_LIBRARY:
				sremquof(A, B, QUO, C);
			break;
			case ALTIVEC_LIBRARY:
				vremquof(A, B, QUO, C);
			break;
		}
	}

/*
[ Power Functions]
                                                                                
    vipowf          Returns x raised to the integer power of y.                 
    vpowf           Returns x raised to the power of y.  Result is more         
                    accurate than using exp(log(x)*y).                          
                                                                                

*/
/*
 *  vipowf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vipowf(float[] A, int[] B, float[] C);
  public static native void  sipowf(float[] A, int[] B, float[] C);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void ipowf(float[] A, int[] B, float[] C) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null || B == null)
					break;
				int size = Math.min(A.length, B.length);
				if (C == null || (C.length < size))
					C = new float[size];
				for (int i = 0; i < size; i++)
					C[i] = (float) Math.pow(A[i], B[i]);
			break;
			case NATIVE_LIBRARY:
				sipowf(A, B, C);
			break;
			case ALTIVEC_LIBRARY:
				vipowf(A, B, C);
			break;
		}
	}

/*
 *  vpowf()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vpowf(float[] A, float[] B, float[] C);
  public static native void  spowf(float[] A, float[] B, float[] C);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void powf(float[] A, float[] B, float[] C) {
		switch (libraryType) {
			case NO_LIBRARY:
				if (A == null || B == null)
					break;
				int size = Math.min(A.length, B.length);
				if (C == null || (C.length < size))
					C = new float[size];
				for (int i = 0; i < size; i++)
					C[i] = (float) Math.pow(A[i], B[i]);
			break;
			case NATIVE_LIBRARY:
				spowf(A, B, C);
			break;
			case ALTIVEC_LIBRARY:
				vpowf(A, B, C);
			break;
		}
	}

/*

    Useful                                                                      

*/
/*
 *  vtablelookup()
 *  
 *  Availability:
 *    Mac OS X:         in version 10.0 and later in vecLib.framework
 *    CarbonLib:        not in Carbon, but vecLib is compatible with Carbon
 *    Non-Carbon CFM:   in vecLib 1.0 and later
 */
  public static native void  vtablelookup(int[] Index_Vect, long[] Table, int[] result);
  public static native void  stablelookup(int[] Index_Vect, long[] Table, int[] result);

// if you call this routine it will just delegate the computation to which method is available
// on the current platform
	public static void tablelookup(int[] Index_Vect, long[] Table, int[] result) {
		switch (libraryType) {
			case NO_LIBRARY:
				System.out.println("Warning: no tablelookup function available in pure java");
/*				if (Index_Vect == null || Table == null)
					break;
				int size = Math.min(Index_Vect.length, Table.length);
				if (result == null || (result.length < size))
					result = new float[size];
				for (int i = 0; i < size; i++)
					result[i] = Math.tablelookup(Index_Vect[i], Table[i]);*/
			break;
			case NATIVE_LIBRARY:
				stablelookup(Index_Vect, Table, result);
			break;
			case ALTIVEC_LIBRARY:
				vtablelookup(Index_Vect, Table, result);
			break;
		}
	}

/*

    Custom functions                                                                      

*/

/*
 *  pseudovoigt() = intensity * ((1 - eta) * Gaussian + eta * Cauchy)
 *
 *  Parameters: intensity - integral value of the PseudoVoigt
 *              eta       - weight between Gaussian and Cauchy component
 *              hwhm      - half width at half maximum
 *              position  - center of the pseudovoigt in the x coordinate
 *              x         - array containing the x coordinate
 *              minindex  - starting index in x to compute the pseudovoigt
 *              maxindex  - final index in x to compute the pseudovoigt
 *              
 *  Compute as:
 *
 *  sqrtln2pi = 4.6971863935E-01;
 *  LN2 = 6.9314718056E-01;
 *
 *  dgx = intensity * (1.0 - eta) * sqrtln2pi / hwhm;
 *  dcx = intensity * eta / (Math.PI * hwhm);
 *	for (int i = minindex, j = 0; i < maxindex; i++, j++) {
 *  	dx = (position - x[i]) / hwhm;
 *    dx *= dx;
 *    if (dx > 1.0E90)
 *      pseudovoigt[j] = 0.0;
 *    else if (dx > 30) {
 *      pseudovoigt[j] = dcx / (1.0 + dx);
 *    } else {
 *      pseudovoigt[j] = dcx / (1.0 + dx) + dgx * Math.exp(- LN2 * dx);
 *    }
 *	}
 *
 *  by Luca Lutterotti  
 */
  public static native float[] vpseudovoigt(float[] x, int minindex, int maxindex, float intensity, 
                                           float eta, float hwhm, float position);
  public static native int vpseudovoigts(float[] x, float[] f, int[] minindex, int[] maxindex, 
  																				float[] intensity, float[] eta, float[] hwhm, float[] position);
  public static native int spseudovoigts(float[] x, float[] f, int[] minindex, int[] maxindex, 
  																				float[] intensity, float[] eta, float[] hwhm, float[] position);
  public static native float[] spseudovoigt(float[] x, int minindex, int maxindex, float intensity, 
                                           float eta, float hwhm, float position);


/*
 *  Angles transformation in texture analysis
 *  The routine compute the transformation returning the angles theta, phi
 *  of the pole figure. The computation is done for an array of points.
 *  The returning theta, phi angles for each point are putted in the 
 *  omega and chi respectively arrays.
 *  All points are for a single reflection located at position (in the theta diffraction
 *  convention).
 *
 *  Parameters: texture_angles_omega 	- omega angle of the goniometer
 *              texture_angles_chi 		- chi angle of the goniometer
 *              texture_angles_phi 		- phi angle of the goniometer
 *              texture_angles_eta 		- eta angle of the goniometer
 *              texture_angles_sampleAngles - omega, chi and phi angles of the sample
 *							position							- position in the diffraction angle unit of the peak
 *																			(in degrees)
 */
 
	public static native int vComputeTextureAngles(	float[] texture_angles_omega,
																									float[] texture_angles_chi,
																									float[] texture_angles_phi,
																									float[] texture_angles_eta,
																									float[] sampleAngles, float position);

	public static native int sComputeTextureAngles(	float[] texture_angles_omega,
																									float[] texture_angles_chi,
																									float[] texture_angles_phi,
																									float[] texture_angles_eta,
																									float[] sampleAngles, float position);

/*

    Test and examples                                                                      

*/


    public JNIAltivec() {
    }

    public static void main (String args[]) {
      System.out.println("Start JNIAltivec example.");
      JNIAltivec altivecExample = new JNIAltivec();
      if (!isAltiVecAvailable())
        System.out.println("No Altivec (processor or library) available, only pure java computation");

				int size = 262144, start = 4, increment = 4;
				for (int i = 4; i <= size; i *= 4)
        	print("Sqrt", i, size, altivecExample.sqrtExample(i, size / i));
				for (int i = 4; i <= size; i *= 4)
        	print("Exp", i, size, altivecExample.expExample(i, size / i));
				for (int i = 4; i <= size; i *= 4)
        	print("Expm1", i, size, altivecExample.expm1Example(i, size / i));
				for (int i = 4; i <= size; i *= 4)
        	print("Log", i, size, altivecExample.logExample(i, size / i));
				for (int i = 4; i <= size; i *= 4)
        	print("Log1p", i, size, altivecExample.log1pExample(i, size / i));
				for (int i = 4; i <= size; i *= 4)
        	print("Cos", i, size, altivecExample.cosExample(i, size / i));
				for (int i = 4; i <= size; i *= 4)
        	print("Tan", i, size, altivecExample.tanExample(i, size / i));
				for (int i = 4; i <= size; i *= 4)
        	print("Acos", i, size, altivecExample.acosExample(i, size / i));
				for (int i = 4; i <= size; i *= 4)
        	print("Atan", i, size, altivecExample.atanExample(i, size / i));
				for (int i = 4; i <= size; i *= 4)
        	print("Divide", i, size, altivecExample.divideExample(i, size / i));
    }
    
    public static void print(String method, int i, int size, long[] time) {
			System.out.println( method + ": size = " + i + ", repeat = " + (size / i) + 
													", Altivec native = " + time[0] + " msecs" +
													", pure native = " + time[1] + " msecs" +
													", pure java = " + time[2] + " msecs"
			);
    }
    
    public long[] divideExample(int repeat, int k100) {
        float[] A1 = {2.5f, 3.1f, 5.6f, -2.4f};
        float[] B1 = {2.4f, 3.2f, 5.5f, 2.3f};
        float[] A = new float[repeat * 4];
        float[] B = new float[repeat * 4];
        float[] one = new float[repeat * 4];
        float[] D = new float[repeat * 4];
        float[] C = new float[repeat * 4];
        int repeat1 = repeat * 4;
        long[] time = new long[3];
        
				// just for hotspot
        int index = 0;
        for (int i = 0; i < repeat1; i++) {
        	A[i] = A1[index];
        	B[i] = B1[index++];
        	if (index == 4)
        		index = 0;
        	one[i] = 1.0f;
        }
				time[0] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						vdivf(A, B, C);
						vdivf(A, C, B);
						vdivf(one, B, D);
						vdivf(C, D, A);
					}
				} catch (Throwable e) {
				}
				time[0] = (System.currentTimeMillis() - time[0]);

/*        index = 0;
        for (int i = 0; i < repeat1; i++) {
        	A[i] = A1[index];
        	B[i] = B1[index++];
        	if (index == 4)
        		index = 0;
        	one[i] = 1.0f;
        }
				time[0] = System.currentTimeMillis();
        for (int k = 0; k < k100; k++) {
        	vdivfF(A, B, C);
        	vdivfF(A, C, B);
        	vdivfF(one, B, D);
        	vdivfF(C, D, A);
        }
				time[0] = (System.currentTimeMillis() - time[0]);*/
//				System.out.println("Time for final full native Altivec computation was: " + time + " millisecs.");
/*        System.out.println("Finished JNIAltivec example. Answer is ");
        for (int i = 0; i < 4; i++)
        	System.out.println(A[i] + " / " + B[i] + " = " + C[i]);*/

        index = 0;
        for (int i = 0; i < repeat1; i++) {
        	A[i] = A1[index];
        	B[i] = B1[index++];
        	if (index == 4)
        		index = 0;
        	one[i] = 1.0f;
        }
				time[1] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						sdivf(A, B, C);
						sdivf(A, C, B);
						sdivf(one, B, D);
						sdivf(C, D, A);
					}
				} catch (Throwable e) {
				}
				time[1] = (System.currentTimeMillis() - time[1]);
/*				System.out.println("Time for final full native computation was: " + time + " millisecs.");
        System.out.println("Finished JNIAltivec example. Answer is ");
        for (int i = 0; i < 4; i++)
        	System.out.println(A[i] + " / " + B[i] + " = " + C[i]);*/

        index = 0;
        for (int i = 0; i < repeat1; i++) {
        	A[i] = A1[index];
        	B[i] = B1[index++];
        	if (index == 4)
        		index = 0;
        	one[i] = 1.0f;
        }
				time[2] = System.currentTimeMillis();
        for (int k = 0; k < k100; k++) {
        	for (int j = 0; j < repeat1; j++) {
        		C[j] = A[j] / B[j];
        		B[j] = A[j] / C[j];
        		D[j] = one[j] / B[j];
        		A[j] = C[j] / D[j];
        	}
				}
				time[2] = (System.currentTimeMillis() - time[2]);
//				System.out.println("Time for final pure java computation was: " + time + " millisecs.");
/*        System.out.println("Finished JNIAltivec example. Answer is ");
        for (int i = 0; i < 4; i++)
        	System.out.println(A[i] + " / " + B[i] + " = " + C[i]);*/
			
				return time;
    }
    
    public long[] sqrtExample(int repeat, int k100) {
        float[] A1 = {2.5f, 3.1f, 5.6f, -2.4f};
        float[] A = new float[repeat * 4];
        float[] C = new float[repeat * 4];
        int repeat1 = repeat * 4;
        long[] time = new long[3];
        
        int index = 0;
        for (int i = 0; i < repeat1; i++) {
        	A[i] = A1[index++];
        	if (index == 4)
        		index = 0;
        }
				time[0] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						vsqrtf(A, C);
					}
				} catch (Throwable e) {
				}
				time[0] = (System.currentTimeMillis() - time[0]);
/*				System.out.println("Time for final full native Altivec computation sqrt was: " + time + " millisecs.");*/

        index = 0;
        for (int i = 0; i < repeat1; i++) {
        	A[i] = A1[index++];
        	if (index == 4)
        		index = 0;
        }
				time[1] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						ssqrtf(A, C);
					}
				} catch (Throwable e) {
				}
				time[1] = (System.currentTimeMillis() - time[1]);
/*				System.out.println("Time for final full native computation sqrt was: " + time + " millisecs.");*/

        index = 0;
        for (int i = 0; i < repeat1; i++) {
        	A[i] = A1[index++];
        	if (index == 4)
        		index = 0;
        }
				time[2] = System.currentTimeMillis();
        for (int k = 0; k < k100; k++)
        	for (int j = 0; j < repeat1; j++) {
        		C[j] = (float) Math.sqrt(A[j]);
        	}
				time[2] = (System.currentTimeMillis() - time[2]);
/*				System.out.println("Time for final pure java computation sqrt was: " + time + " millisecs.");*/

				return time;
    }
    
    public long[] expExample(int repeat, int k100) {
        float[] A1 = {2.5f, 3.1f, 5.6f, -2.4f};
        float[] A = new float[repeat * 4];
        float[] C = new float[repeat * 4];
        int repeat1 = repeat * 4;
        long[] time = new long[3];
        
        int index = 0;
        for (int i = 0; i < repeat1; i++) {
        	A[i] = A1[index++];
        	if (index == 4)
        		index = 0;
        }

				time[0] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						vexpf(A, C);
					}
				} catch (Throwable e) {
				}
				time[0] = (System.currentTimeMillis() - time[0]);
/*				System.out.println("Time for final full native Altivec computation exp was: " + time + " millisecs.");*/

				time[1] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						sexpf(A, C);
					}
				} catch (Throwable e) {
				}
				time[1] = (System.currentTimeMillis() - time[1]);
/*				System.out.println("Time for final full native computation exp was: " + time + " millisecs.");*/

				time[2] = System.currentTimeMillis();
        for (int k = 0; k < k100; k++)
        	for (int j = 0; j < repeat1; j++) {
        		C[j] = (float) Math.exp(A[j]);
        	}
				time[2] = (System.currentTimeMillis() - time[2]);
/*				System.out.println("Time for final pure java computation exp was: " + time + " millisecs.");*/

				return time;
    }
    
    public long[] expm1Example(int repeat, int k100) {
        float[] A1 = {2.5f, 3.1f, 5.6f, -2.4f};
        float[] A = new float[repeat * 4];
        float[] C = new float[repeat * 4];
        int repeat1 = repeat * 4;
        long[] time = new long[3];
        
        int index = 0;
        for (int i = 0; i < repeat1; i++) {
        	A[i] = A1[index++];
        	if (index == 4)
        		index = 0;
        }

				time[0] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						vexpm1f(A, C);
					}
				} catch (Throwable e) {
				}
				time[0] = (System.currentTimeMillis() - time[0]);
/*				System.out.println("Time for final full native Altivec computation expm1 was: " + time + " millisecs.");*/

				time[1] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						sexpm1f(A, C);
					}
				} catch (Throwable e) {
				}
				time[1] = (System.currentTimeMillis() - time[1]);
/*				System.out.println("Time for final full native computation expm1 was: " + time + " millisecs.");*/

				time[2] = System.currentTimeMillis();
        for (int k = 0; k < k100; k++)
        	for (int j = 0; j < repeat1; j++) {
        		C[j] = (float) (Math.exp(A[j]) - 1.0);
        	}
				time[2] = (System.currentTimeMillis() - time[2]);
/*				System.out.println("Time for final pure java computation expm1 was: " + time + " millisecs.");*/

				return time;
    }
    
    public long[] logExample(int repeat, int k100) {
        float[] A1 = {2.5f, 3.1f, 5.6f, -2.4f};
        float[] A = new float[repeat * 4];
        float[] C = new float[repeat * 4];
        int repeat1 = repeat * 4;
        long[] time = new long[3];
        
        int index = 0;
        for (int i = 0; i < repeat1; i++) {
        	A[i] = A1[index++];
        	if (index == 4)
        		index = 0;
        }

				time[0] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						vlogf(A, C);
					}
				} catch (Throwable e) {
				}
				time[0] = (System.currentTimeMillis() - time[0]);
/*				System.out.println("Time for final full native Altivec computation log was: " + time + " millisecs.");*/

				time[1] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						slogf(A, C);
					}
				} catch (Throwable e) {
				}
				time[1] = (System.currentTimeMillis() - time[1]);
/*				System.out.println("Time for final full native computation log was: " + time + " millisecs.");*/

				time[2] = System.currentTimeMillis();
        for (int k = 0; k < k100; k++)
        	for (int j = 0; j < repeat1; j++) {
        		C[j] = (float) Math.log(A[j]);
        	}
				time[2] = (System.currentTimeMillis() - time[2]);
/*				System.out.println("Time for final pure java computation log was: " + time + " millisecs.");*/

				return time;
    }
    
    public long[] log1pExample(int repeat, int k100) {
        float[] A1 = {2.5f, 3.1f, 5.6f, -2.4f};
        float[] A = new float[repeat * 4];
        float[] C = new float[repeat * 4];
        int repeat1 = repeat * 4;
        long[] time = new long[3];
        
        int index = 0;
        for (int i = 0; i < repeat1; i++) {
        	A[i] = A1[index++];
        	if (index == 4)
        		index = 0;
        }

				time[0] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						vlog1pf(A, C);
					}
				} catch (Throwable e) {
				}
				time[0] = (System.currentTimeMillis() - time[0]);
/*				System.out.println("Time for final full native Altivec computation log1p was: " + time + " millisecs.");*/

				time[1] = System.currentTimeMillis();
        for (int k = 0; k < k100; k++) {
        	slog1pf(A, C);
        }
				time[1] = (System.currentTimeMillis() - time[1]);
/*				System.out.println("Time for final full native computation log1p was: " + time + " millisecs.");*/

				time[2] = System.currentTimeMillis();
        for (int k = 0; k < k100; k++)
        	for (int j = 0; j < repeat1; j++) {
        		C[j] = (float) Math.log(1.0 + A[j]);
        	}
				time[2] = (System.currentTimeMillis() - time[2]);
/*				System.out.println("Time for final pure java computation log1p was: " + time + " millisecs.");*/

				return time;
    }
    
    public long[] cosExample(int repeat, int k100) {
        float[] A1 = {2.5f, 3.1f, 5.6f, -2.4f};
        float[] A = new float[repeat * 4];
        float[] C = new float[repeat * 4];
        int repeat1 = repeat * 4;
        long[] time = new long[3];
        
        int index = 0;
        for (int i = 0; i < repeat1; i++) {
        	A[i] = A1[index++];
        	if (index == 4)
        		index = 0;
        }

				time[0] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						vcosf(A, C);
					}
				} catch (Throwable e) {
				}
				time[0] = (System.currentTimeMillis() - time[0]);
/*				System.out.println("Time for final full native Altivec computation cos was: " + time + " millisecs.");*/

				time[1] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						scosf(A, C);
					}
				} catch (Throwable e) {
				}
				time[1] = (System.currentTimeMillis() - time[1]);
/*				System.out.println("Time for final full native computation cos was: " + time + " millisecs.");*/

				time[2] = System.currentTimeMillis();
        for (int k = 0; k < k100; k++)
        	for (int j = 0; j < repeat1; j++) {
        		C[j] = (float) Math.cos(A[j]);
        	}
				time[2] = (System.currentTimeMillis() - time[2]);
/*				System.out.println("Time for final pure java computation cos was: " + time + " millisecs.");*/

				return time;
    }
    
    public long[] tanExample(int repeat, int k100) {
        float[] A1 = {2.5f, 3.1f, 5.6f, -2.4f};
        float[] A = new float[repeat * 4];
        float[] C = new float[repeat * 4];
        int repeat1 = repeat * 4;
        long[] time = new long[3];
        
        int index = 0;
        for (int i = 0; i < repeat1; i++) {
        	A[i] = A1[index++];
        	if (index == 4)
        		index = 0;
        }

				time[0] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						vtanf(A, C);
					}
				} catch (Throwable e) {
				}
				time[0] = (System.currentTimeMillis() - time[0]);
/*				System.out.println("Time for final full native Altivec computation tan was: " + time + " millisecs.");*/

				time[1] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						stanf(A, C);
					}
				} catch (Throwable e) {
				}
				time[1] = (System.currentTimeMillis() - time[1]);
/*				System.out.println("Time for final full native computation tan was: " + time + " millisecs.");*/

				time[2] = System.currentTimeMillis();
        for (int k = 0; k < k100; k++)
        	for (int j = 0; j < repeat1; j++) {
        		C[j] = (float) Math.tan(A[j]);
        	}
				time[2] = (System.currentTimeMillis() - time[2]);
/*				System.out.println("Time for final pure java computation tan was: " + time + " millisecs.");*/

				return time;
    }
    
    public long[] acosExample(int repeat, int k100) {
        float[] A1 = {2.5f, 3.1f, 5.6f, -2.4f};
        float[] A = new float[repeat * 4];
        float[] C = new float[repeat * 4];
        int repeat1 = repeat * 4;
        long[] time = new long[3];
        
        int index = 0;
        for (int i = 0; i < repeat1; i++) {
        	A[i] = A1[index++];
        	if (index == 4)
        		index = 0;
        }

				time[0] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						vacosf(A, C);
					}
				} catch (Throwable e) {
				}
				time[0] = (System.currentTimeMillis() - time[0]);
/*				System.out.println("Time for final full native Altivec computation acos was: " + time + " millisecs.");*/

				time[1] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						sacosf(A, C);
					}
				} catch (Throwable e) {
				}
				time[1] = (System.currentTimeMillis() - time[1]);
/*				System.out.println("Time for final full native computation acos was: " + time + " millisecs.");*/

				time[2] = System.currentTimeMillis();
        for (int k = 0; k < k100; k++)
        	for (int j = 0; j < repeat1; j++) {
        		C[j] = (float) Math.acos(A[j]);
        	}
				time[2] = (System.currentTimeMillis() - time[2]);
/*				System.out.println("Time for final pure java computation acos was: " + time + " millisecs.");*/

				return time;
    }
    
    public long[] atanExample(int repeat, int k100) {
        float[] A1 = {2.5f, 3.1f, 5.6f, -2.4f};
        float[] A = new float[repeat * 4];
        float[] C = new float[repeat * 4];
        int repeat1 = repeat * 4;
        long[] time = new long[3];
        
        int index = 0;
        for (int i = 0; i < repeat1; i++) {
        	A[i] = A1[index++];
        	if (index == 4)
        		index = 0;
        }

				time[0] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						vatanf(A, C);
					}
				} catch (Throwable e) {
				}
				time[0] = (System.currentTimeMillis() - time[0]);
/*				System.out.println("Time for final full native Altivec computation atan was: " + time + " millisecs.");*/

				time[1] = System.currentTimeMillis();
				try {
					for (int k = 0; k < k100; k++) {
						satanf(A, C);
					}
				} catch (Throwable e) {
				}
				time[1] = (System.currentTimeMillis() - time[1]);
/*				System.out.println("Time for final full native computation atan was: " + time + " millisecs.");*/

				time[2] = System.currentTimeMillis();
        for (int k = 0; k < k100; k++)
        	for (int j = 0; j < repeat1; j++) {
        		C[j] = (float) Math.atan(A[j]);
        	}
				time[2] = (System.currentTimeMillis() - time[2]);
/*				System.out.println("Time for final pure java computation atan was: " + time + " millisecs.");*/

				return time;
    }
    
}



